@@ -19,9 +19,9 @@
 ;; URL:      https://github.com/meedstrom/org-mem
 ;; Created:  2025-03-15
 ;; Keywords: text
-;; Package-Version: 0.23.2
-;; Package-Revision: 0.23.2-0-g521864ca617a
-;; Package-Requires: ((emacs "29.1") (el-job "2.4.8") (llama "0.5.0"))
+;; Package-Version: 0.25.0
+;; Package-Revision: 0.25.0-0-g531a6b571ba0
+;; Package-Requires: ((emacs "29.1") (el-job "2.5.1") (llama "0.5.0"))
 
 ;;; Commentary:
 
@@ -60,6 +60,9 @@
 (declare-function org-id-alist-to-hash "org-id")
 (declare-function org-id-hash-to-alist "org-id")
 (define-obsolete-variable-alias 'org-mem-watch-dirs-exclude 'org-mem-exclude "0.13.0 (May 2025)")
+(unless (and (boundp 'el-job-old-internal-version)
+             (>= el-job-old-internal-version 100))
+  (error "Update to el-job 2.5.1+ to use this version of org-mem"))
 
 (defgroup org-mem nil "Fast info from a large amount of Org file contents."
   :group 'org)
@@ -179,6 +182,11 @@ process, because `load-history' is used to find the corresponding file."
   :type '(alist :key-type symbol :value-type sexp)
   :package-version '(org-mem . "0.21.0"))
 
+(defcustom org-mem-eval-forms nil
+  "Experimental.
+Quoted forms to eval in child process."
+  :type '(repeat sexp))
+
 (defcustom org-mem-ignore-regions-regexps
   '(("^[ \t]*:ROAM_REFS:" . "$")
     ("^[ \t]*:BACKLINKS:" . "$")
@@ -200,7 +208,6 @@ These are hooks called many times:
 - `org-mem-record-link-functions'
 - `org-mem-forget-file-functions'
 - `org-mem-forget-entry-functions'
-- `org-mem-forget-link-functions'
 
 This lets a function on these hooks sidestep the performance overhead of
 `with-temp-buffer' or `with-work-buffer', in favor of using the
@@ -246,6 +253,10 @@ in `org-link-plain-re').  In practice, it is often an org-id like
 
 NOTE! A future version may omit the sigil @ in citekeys.")
 
+(defvar org-mem--target<>old-links (make-hash-table :test 'equal)
+  "Previous state of `org-mem--target<>links'.
+For downstream use.")
+
 (defvar org-mem--internal-entry-id<>links (make-hash-table :test 'eq)
   "1:N table mapping internal entry ID to list of `org-mem-link' records.
 The list represents all links found in that entry,
@@ -294,6 +305,7 @@ Note: All tables cleared often, meant for memoizations."
   (description        nil :read-only t :type string-or-nil)
   (citation-p         nil :read-only t :type boolean)
   (nearby-id          nil :read-only t :type string-or-nil)
+  (supplement         nil :read-only t :type string-or-nil)
   (-internal-entry-id -1  :read-only t :type integer))
 
 (cl-defstruct org-mem-entry
@@ -334,8 +346,7 @@ perfectly with `org-id-locations'."
 (defun org-mem-all-files-expanded ()
   "All Org files, with tilde expansion applied."
   (with-memoization (org-mem--table 0 'org-mem-all-files-expanded)
-    (with-temp-buffer ;; No buffer-env
-      (mapcar #'expand-file-name (org-mem-all-files)))))
+    (mapcar #'directory-abbrev-apply (org-mem-all-file-truenames))))
 
 (defun org-mem-all-files ()
   "All Org files."
@@ -619,7 +630,7 @@ use `org-mem-entry-olpath-with-file-title-or-basename' instead."
           (pop olp))
         olp))))
 
-;; Better than above: single-argument func.
+;; Better than above; unary function.
 (defun org-mem-entry-olpath-with-file-title-or-basename (entry)
   "Outline path to ENTRY, including file #+title.
 Use file basename if there is no #+title."
@@ -1192,9 +1203,6 @@ hence the name.  Contrast `org-mem-post-full-scan-functions'.")
 (defvar org-mem-forget-entry-functions nil
   "Hook passed one forgotten `org-mem-entry' object.")
 
-(defvar org-mem-forget-link-functions nil
-  "Hook passed one forgotten `org-mem-link' object.")
-
 (defvar org-mem--problems nil)
 (defvar org-mem--title-collisions nil)
 (defvar org-mem--id-collisions nil)
@@ -1236,6 +1244,7 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
                        :if-busy 'takeover
                        :inject-vars (append (org-mem--mk-work-vars) org-mem-inject-vars)
                        :load-features (append '(org-mem-parser) org-mem-load-features)
+                       :eval org-mem-eval-forms
                        :inputs files
                        :funcall-per-input #'org-mem-parser--parse-file
                        :callback #'org-mem--finalize-full-scan)
@@ -1275,18 +1284,22 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
   (clrhash org-mem--truename<>metadata)
   (clrhash org-mem--truename<>entries)
   (clrhash org-mem--internal-entry-id<>links)
-  (clrhash org-mem--target<>links)
   (setq org-mem--title-collisions nil)
   (seq-let (bad-paths file-data entries links problems) parse-results
-    (when bad-paths
-      (org-mem--invalidate-file-names bad-paths))
+    (org-mem--invalidate-file-names bad-paths)
     (with-current-buffer
         (setq org-mem-scratch (get-buffer-create " *org-mem-scratch*" t))
       (dolist (fdata file-data)
         (puthash (car fdata) fdata org-mem--truename<>metadata)
         (run-hook-with-args 'org-mem-record-file-functions fdata))
-      (mapc #'org-mem--record-entry entries)
-      (mapc #'org-mem--record-link links))
+      (dolist (entry entries)
+        (org-mem--record-entry entry)
+        (run-hook-with-args 'org-mem-record-entry-functions entry))
+      (dolist (link links)
+        (push link (gethash (org-mem-link--internal-entry-id link)
+                            org-mem--internal-entry-id<>links))
+        (run-hook-with-args 'org-mem-record-link-functions link)))
+    (org-mem--rebuild-specially-indexed-tables)
 
     (setq org-mem--time-elapsed
           (float-time (time-since org-mem--time-at-begin-full-scan)))
@@ -1327,13 +1340,6 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
       (setq org-mem--problems (append problems org-mem--problems))
       (message "Scan had problems, see M-x org-mem-list-problems"))))
 
-(defun org-mem--record-link (link)
-  "Add info related to LINK to various tables."
-  (push link (gethash (org-mem-link-target link) org-mem--target<>links))
-  (push link (gethash (org-mem-link--internal-entry-id link)
-                      org-mem--internal-entry-id<>links))
-  (run-hook-with-args 'org-mem-record-link-functions link))
-
 (defun org-mem--record-entry (entry)
   "Add info related to ENTRY to various tables."
   (let ((id (org-mem-entry-id entry))
@@ -1344,6 +1350,7 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
     (push entry (gethash truename org-mem--truename<>entries))
     (when id
       (org-mem--maybe-snitch-to-org-id entry)
+      (puthash id entry org-mem--id<>entry)
       (when title
         (let ((other-id (gethash title org-mem--title<>id)))
           (when (and other-id (not (string= id other-id)))
@@ -1352,9 +1359,21 @@ overrides a default message printed when `org-mem-do-cache-text' is t."
         ;; REVIEW: Should we fallback on file name, i.e. use
         ;; `org-mem-entry-title' instead of `org-mem-entry-title-maybe'?
         ;; It would affect function `org-mem-id-by-title'.
-        (puthash title id org-mem--title<>id))
-      (puthash id entry org-mem--id<>entry))
-    (run-hook-with-args 'org-mem-record-entry-functions entry)))
+        (puthash title id org-mem--title<>id)))))
+
+(defun org-mem--rebuild-specially-indexed-tables ()
+  "Rebuild table `org-mem--target<>links'.
+
+A \"specially indexed\" table is one that cannot be trivially updated when
+one file is re-scanned, since a given key in it may list values from
+multiple files.  So rather than try to find the particular values to
+remove, it's easiest to wipe and re-build."
+  (setq org-mem--target<>old-links (copy-hash-table org-mem--target<>links))
+  (clrhash org-mem--target<>links)
+  (cl-loop
+   for links being each hash-value of org-mem--internal-entry-id<>links
+   do (dolist (link links)
+        (push link (gethash (org-mem-link-target link) org-mem--target<>links)))))
 
 (defun org-mem--maybe-snitch-to-org-id (entry)
   "Add applicable ENTRY data to `org-id-locations'.
@@ -1456,7 +1475,12 @@ changing the meaning of \"~\" or \"~USER\", or runtime changes to
       (and (stringp wild-file)
            (not (file-remote-p wild-file))
            (let (file-name-handler-alist)
-             (cl-assert (file-name-absolute-p wild-file))
+             (unless (file-name-absolute-p wild-file)
+               ;; NOTE: For a lot of users, the above condition always comes
+               ;; true, but it depends on e.g. which completion system you use
+               ;; with a command like `rename-file'.
+               ;; https://github.com/meedstrom/org-mem/issues/30
+               (setq wild-file (expand-file-name wild-file)))
              (if (file-exists-p wild-file)
                  (let* ((truename (file-truename wild-file))
                         (abbr-true (org-mem--fast-abbrev truename)))
@@ -1468,22 +1492,23 @@ changing the meaning of \"~\" or \"~USER\", or runtime changes to
   "Scrub bad file names BAD in the tables that can pollute a reset.
 Notably, invalidate part of the cache used by `org-mem--truename-maybe'.
 If `org-mem-do-sync-with-org-id' t, also scrub `org-id-locations'."
-  ;; Example situation: File WILD is a symlink that changed destination.
-  ;; So cached TRUE led to a nonexistent file in the last scan.
-  ;; Now invalidate it so we cache a correct TRUE next time.
-  (maphash (lambda (wild true)
-             (when (member true bad)
-               (push wild bad)))
-           org-mem--wild-filename<>truename)
-  (dolist (bad bad)
-    (remhash bad org-mem--wild-filename<>truename))
-  (when (and org-mem-do-sync-with-org-id
-             (org-mem--try-ensure-org-id-table-p))
-    (setq org-id-locations
-          (org-id-alist-to-hash
-           (cl-loop for cell in (org-id-hash-to-alist org-id-locations)
-                    unless (member (car cell) bad)
-                    collect cell)))))
+  (when bad
+    ;; Example situation: File WILD is a symlink that changed destination.
+    ;; So cached TRUE led to a nonexistent file in the last scan.
+    ;; Now invalidate it so we cache a correct TRUE next time.
+    (maphash (lambda (wild true)
+               (when (member true bad)
+                 (push wild bad)))
+             org-mem--wild-filename<>truename)
+    (dolist (bad bad)
+      (remhash bad org-mem--wild-filename<>truename))
+    (when (and org-mem-do-sync-with-org-id
+               (org-mem--try-ensure-org-id-table-p))
+      (setq org-id-locations
+            (org-id-alist-to-hash
+             (cl-loop for cell in (org-id-hash-to-alist org-id-locations)
+                      unless (member (car cell) bad)
+                      collect cell))))))
 
 (defun org-mem--try-ensure-org-id-table-p ()
   "Coerce `org-id-locations' into hash table form, return nil on fail."
@@ -1763,7 +1788,7 @@ org-id-locations:
       (org-mem--scan-full))))
 
 
-(defvar org-mem--bump-int 17 "Not a version number, but bumped sometimes.")
+(defvar org-mem--bump-int 19 "Not a version number, but bumped sometimes.")
 (defmacro org-mem--def-whiny-alias (old new when removed-by)
   "Define function OLD as effectively an alias for NEW.
 Also, calling OLD will emit a deprecation warning the first time.
@@ -1798,6 +1823,7 @@ may be removed from the package."
 
 (define-obsolete-function-alias 'org-mem-entry-olpath-with-file-title-with-self #'org-mem-entry-olpath-with-self-with-file-title  "0.22.0 (2025-10-01)")
 (define-obsolete-function-alias 'org-mem-olpath-with-file-title-with-self       #'org-mem-olpath-with-self-with-file-title        "0.22.0 (2025-10-01)")
+(defvar org-mem-forget-link-functions :obsolete) ;; 2025-10-15 but unused for long already
 
 (provide 'org-mem)
 
